@@ -1,7 +1,7 @@
 import typing as t
-import logging
 import os
 import sys
+import enum
 
 import gi
 
@@ -25,22 +25,49 @@ from .click import (
 from . import GST_PLUGIN_PATH
 
 
-logger = logging.getLogger(__name__)
-logger.setLevel(logging.INFO)
+RGB_FORMATS = [
+    "RGB",
+
+    "RGBx",
+    "xRGB",
+
+    "RGBA",
+    "ARGB",
+
+    "BGR",
+
+    "BGRx",
+    "xBGR",
+
+    "BGRA",
+    "ABGR",
+]
 
 
-# hack to let Gst.ElementFactory.make("...") to find custom plugins
-os.environ["GST_PLUGIN_PATH"] = ":".join(
-    [GST_PLUGIN_PATH] + os.environ.get("GST_PLUGIN_PATH", "").split(":")
-)
+class HardwareAccelAPI(enum.Enum):
+    off = enum.auto()
+    vaapi = enum.auto()
 
-# hack to ensure interperter path is on PATH so that custom plugins will use the correct interpeter
-os.environ["PATH"] = ":".join(
-    [str(Path(sys.executable).parent)] + os.environ.get("PATH", "").split(":")
-)
 
-GObject.threads_init()
-Gst.init(None)
+def init():
+    # hack to let Gst.ElementFactory.make("...") to find custom plugins
+    os.environ["GST_PLUGIN_PATH"] = ":".join(
+        [GST_PLUGIN_PATH] + os.environ.get("GST_PLUGIN_PATH", "").split(":")
+    )
+
+    # hack to ensure interperter path is on PATH so that custom plugins will use the correct interpeter
+    os.environ["PATH"] = ":".join(
+        [str(Path(sys.executable).parent)] + os.environ.get("PATH", "").split(":")
+    )
+
+    # TODO: Find out why this is needed. On some systems (e.g. nixos) forking
+    # will cause Python plugins not to be loaded properly. No real harm in
+    # disabling it AFAIK.
+    os.environ["GST_REGISTRY_FORK"] = "no"
+
+    GObject.threads_init()
+    if not Gst.init_check(None):
+        raise RuntimeError("failed to initialize gstreamer")
 
 
 def add_filters(
@@ -53,10 +80,14 @@ def add_filters(
     background_blur: t.Optional[int],
     selfie_segmentation_model: SelfieSegmentationModel,
     selfie_segmentation_threshold: int,
+    hw_accel_api: HardwareAccelAPI,
+    verbose: bool,
 ) -> None:
     """
     Run filters pipeline.
     """
+    init()
+
     caps = query_device_caps(input_dev)
 
     if caps is None:
@@ -98,19 +129,44 @@ def add_filters(
     new_caps.append_structure(s)
 
     click.echo(
-        f"Selectd input: media-type={s.get_name()}, width={s.get_value('width')} "
+        f"Selected input: media-type={s.get_name()}, width={s.get_value('width')} "
         f"height={s.get_value('height')} framerate={s.get_value('framerate')}"
     )
 
+    enable_hwdec_plugins(hw_accel_api)
+
     pipeline = Gst.Pipeline.new()
 
     src = Gst.ElementFactory.make("v4l2src")
     inputfilter = Gst.ElementFactory.make("capsfilter")
-    decodebin = Gst.ElementFactory.make("decodebin")
-    rgbconvert = Gst.ElementFactory.make("videoconvert")
+    decodebin = Gst.ElementFactory.make("decodebin3")
+
+    if hw_accel_api == HardwareAccelAPI.vaapi:
+        # vaapipostproc doesn't seem to support going directly to RGB on some
+        # hardware. So prefer direct RGB conversion if possible, otherwise
+        # fallback to some variant of RGBA and then use videoconvert to
+        # to get to RGB.
+        c= ";".join(f"video/x-raw, format={f}" for f in reversed(RGB_FORMATS))
+        rgbconvert = Gst.parse_bin_from_description(
+            f"capsfilter caps=video/x-raw(memory:VASurface) ! "
+            f"vaapipostproc ! {c} ! videoconvert",
+            True
+        )
+    else:
+        rgbconvert = Gst.ElementFactory.make("videoconvert")
+
     rgbfilter = Gst.ElementFactory.make("capsfilter")
     tee = Gst.ElementFactory.make("tee")
-    sinkconvert = Gst.ElementFactory.make("videoconvert")
+
+    if hw_accel_api == HardwareAccelAPI.vaapi:
+        c = ";".join(f"video/x-raw, format={f}" for f in RGB_FORMATS)
+        sinkconvert = Gst.parse_bin_from_description(
+            f"videoconvert ! {c} ! vaapipostproc",
+            True
+        )
+    else:
+        sinkconvert = Gst.ElementFactory.make("videoconvert")
+
     sinkfilter = Gst.ElementFactory.make("capsfilter")
     sink = Gst.ElementFactory.make("v4l2sink")
 
@@ -186,9 +242,13 @@ def add_filters(
 
 
     loop = GObject.MainLoop()
+
+    if verbose:
+        pipeline.add_property_deep_notify_watch(None, True)
+
     bus = pipeline.get_bus()
     bus.add_signal_watch()
-    bus.connect("message", on_bus_message, loop)
+    bus.connect("message", on_bus_message, loop, verbose, pipeline)
 
     pipeline.set_state(Gst.State.PLAYING)
     try:
@@ -202,19 +262,58 @@ def add_filters(
 def on_bus_message(
     bus: Gst.Bus,
     message: Gst.Message,
-    loop: GObject.MainLoop
+    loop: GObject.MainLoop,
+    verbose: bool,
+    pipeline: Gst.Pipeline,
 ) -> bool:
     mtype = message.type
+    src = message.src
+    src_path = src.get_path_string()
+
+    if verbose:
+        click.echo(f"\nsrc: {src_path}")
+        click.echo(f"type: {mtype}")
 
     if mtype == Gst.MessageType.EOS:
         loop.quit()
+
     elif mtype == Gst.MessageType.ERROR:
-        logger.error("%s: %s", *message.parse_error())
+        gerror, debug = message.parse_error()
+        click.echo(f"Error from {src_path}: {gerror.message}")
+        if verbose:
+            click.echo(f"Debug: {debug}")
+
         loop.quit()
+
     elif mtype == Gst.MessageType.WARNING:
-        logger.warning("%s: %s", *message.parse_warning())
+        gerror, debug = message.parse_warning()
+        click.echo(f"Warning from {src_path}: {gerror.message}")
+        if verbose:
+            click.echo(f"Debug: {debug}")
+
     elif mtype == Gst.MessageType.INFO:
-        logger.info("%s: %s", *message.parse_info())
+        gerror, debug = message.parse_info()
+        click.echo(f"Info from {src_path}: {gerror.message}")
+        if verbose:
+            click.echo(f"Debug: {debug}")
+
+    elif mtype == Gst.MessageType.PROPERTY_NOTIFY and verbose:
+        obj, name, value = message.parse_property_notify()
+
+        click.echo(f"{name}: {value}")
+
+    elif mtype == Gst.MessageType.STATE_CHANGED and src == pipeline:
+        old, new, pending = message.parse_state_changed()
+
+        click.echo("Pipeline: ", nl=False)
+        if new == Gst.State.PAUSED:
+            click.echo("PAUSED")
+        elif new == Gst.State.READY:
+            click.echo("READY")
+        elif new == Gst.State.PLAYING:
+            click.echo("RUNNING")
+        elif new == Gst.State.NULL:
+            click.echo("NULL")
 
     return True
 
@@ -247,6 +346,8 @@ def print_device_caps(
     """
     Print device capabilities and exit.
     """
+    init()
+
     if not value or ctx.resilient_parsing:
         return
 
@@ -275,3 +376,29 @@ def print_device_caps(
     console.print(table)
 
     ctx.exit(0)
+
+
+def enable_hwdec_plugins(api: HardwareAccelAPI) -> None:
+    if api == HardwareAccelAPI.off:
+        return
+
+    reg = Gst.Registry.get()
+
+    if api == HardwareAccelAPI.vaapi:
+        plugins = [
+            "vaapijpegdec",
+            "vaapimpeg2dec",
+            "vaapih264dec",
+            "vaapih265dec",
+            "vaapivc1dec",
+        ]
+
+        for p in plugins:
+            factory = Gst.ElementFactory.find(p)
+
+            if factory is None:
+                continue
+
+            factory.set_rank(Gst.Rank.PRIMARY + 1)
+
+            reg.add_feature(factory)
